/*
 * Copyright 2011 Nicolas Limare <nicolas.limare@cmla.ens-cachan.fr>
 *
 * Copying and distribution of this file, with or without
 * modification, are permitted in any medium without royalty provided
 * the copyright notice and this notice are preserved.  This file is
 * offered as-is, without any warranty.
 */

#include <stdlib.h>
#include <limits.h>
#include <stdio.h>
#include <math.h>

#include "cpucycles.h"

#define TEST_SIZE 1024
#ifndef BLOCK_SIZE
#define BLOCK_SIZE 1
#endif

static int cmp(const void *a, const void *b)
{
    return (int) (*(const long long int *) a - *(const long long int *) b);
}

static long long int median_cycles(long long int *cycles, long long int start)
{
    size_t i;

    for (i=0; i<TEST_SIZE; i++)
	cycles[i] -= start;
    for (i=TEST_SIZE-1; i>0; i--)
	cycles[i] -= cycles[i - 1];
    qsort((void *) cycles, TEST_SIZE, sizeof(long long int), &cmp);
    return cycles[TEST_SIZE / 2];
}

int main()
{
    unsigned char r[TEST_SIZE], g[TEST_SIZE], b[TEST_SIZE];
    size_t i, j;
    long long int start, cycles[TEST_SIZE];
    int tmp;
    /*
     * the volatile keyword avoids the removal of unused variables
     * by the compiler optimization
     */
    volatile unsigned char y[TEST_SIZE];

    /* initialization from pseudo-random numbers */
    srand(0);
    for (i=0; i<TEST_SIZE; i++) {
	r[i] = (unsigned char) (rand() % (UCHAR_MAX + 1));
	g[i] = (unsigned char) (rand() % (UCHAR_MAX + 1));
	b[i] = (unsigned char) (rand() % (UCHAR_MAX + 1));
    }

    /* floating-point * +, floor rounded */
    start = cpucycles();
    for (i=0; i<TEST_SIZE; i++) {
	for (j=0; j<BLOCK_SIZE; j++)
	    y[i] = (unsigned char)
		floor(.212671 * r[i] + .715160 * g[i] + .072169 * b[i] + .5);
	cycles[i] = cpucycles();
    }
    printf("%lli cycles\tfloating-point * +, floor rounded\n",
	   median_cycles(cycles, start));

    /* floating-point * +, cast rounded */
    start = cpucycles();
    for (i=0; i<TEST_SIZE; i++) {
	for (j=0; j<BLOCK_SIZE; j++)
	    y[i] = (unsigned char)
		(.212671 * r[i] + .715160 * g[i] + .072169 * b[i] + .5);
	cycles[i] = cpucycles();
    }
    printf("%lli cycles\tfloating-point * +, cast rounded\n",
	   median_cycles(cycles, start));

    /* floating-point * +, cast truncated */
    start = cpucycles();
    for (i=0; i<TEST_SIZE; i++) {
	for (j=0; j<BLOCK_SIZE; j++)
	    y[i] = (unsigned char)
		(.212671 * r[i] + .715160 * g[i] + .072169 * b[i]);
	cycles[i] = cpucycles();
    }
    printf("%lli cycles\tfloating-point * +, cast truncated\n",
	   median_cycles(cycles, start));

    /* integer * + /, division truncated */
    start = cpucycles();
    for (i=0; i<TEST_SIZE; i++) {
	for (j=0; j<BLOCK_SIZE; j++)
	    y[i] = (unsigned char)
		((6969 * r[i] + 23434 * g[i] + 2365 * b[i]) / 32768);
	cycles[i] = cpucycles();
    }
    printf("%lli cycles\tinteger * + /, division truncated\n",
	   median_cycles(cycles, start));

    /* integer * + floating-point /, floor rounded */
    start = cpucycles();
    for (i=0; i<TEST_SIZE; i++) {
	for (j=0; j<BLOCK_SIZE; j++)
	    y[i] = (unsigned char)
		floor((6969 * r[i] + 23434 * g[i] + 2365 * b[i])
		      / 32768. + .5);
	cycles[i] = cpucycles();
    }
    printf("%lli cycles\tinteger * + floating-point /, floor rounded\n",
	   median_cycles(cycles, start));

    /* integer * + floating-point /, cast rounded */
    start = cpucycles();
    for (i=0; i<TEST_SIZE; i++) {
	for (j=0; j<BLOCK_SIZE; j++)
	    y[i] = (unsigned char)
		((6969 * r[i] + 23434 * g[i] + 2365 * b[i]) / 32768. + .5);
	cycles[i] = cpucycles();
    }
    printf("%lli cycles\tinteger * + floating-point /, cast rounded\n",
	   median_cycles(cycles, start));

    /* integer * + >>, shift truncated */
    start = cpucycles();
    for (i=0; i<TEST_SIZE; i++) {
	for (j=0; j<BLOCK_SIZE; j++)
	    y[i] = (unsigned char)
		(6969 * r[i] + 23434 * g[i] + 2365 * b[i]) >> 15;
	cycles[i] = cpucycles();
    }
    printf("%lli cycles\tinteger * + >>, shift truncated\n",
	   median_cycles(cycles, start));

    /* integer * + >>, shift rounded */
    start = cpucycles();
    for (i=0; i<TEST_SIZE; i++) {
	for (j=0; j<BLOCK_SIZE; j++)
	    y[i] = (unsigned char)
		(((6969 * r[i] + 23434 * g[i] + 2365 * b[i]) >> 15)
		 + (((6969 * r[i] + 23434 * g[i] + 2365 * b[i])
		     & (1 << 14)) >> 14));
	cycles[i] = cpucycles();
    }
    printf("%lli cycles\tinteger * + >>, shift rounded\n",
	   median_cycles(cycles, start));

    /* integer * + >>, shift rounded with tmp */
    start = cpucycles();
    for (i=0; i<TEST_SIZE; i++) {
	for (j=0; j<BLOCK_SIZE; j++) {
	    tmp = 6969 * r[i] + 23434 * g[i] + 2365 * b[i];
	    y[i] = (unsigned char)
		((tmp >> 15) + ((tmp & (1 << 14)) >> 14));
	}
	cycles[i] = cpucycles();
    }
    printf("%lli cycles\tinteger * + >>, shift rounded with tmp\n",
	   median_cycles(cycles, start));

    return EXIT_SUCCESS;
}
